import math
import copy

import numpy as np

from . import equations

class Ray:
    def __init__(self, y=0, u=0, n=1, λ0=532e-9):
        self.y = y
        self.u = u
        self.n = n
        self.λ0 = λ0

    def __repr__(self):
        return f"[[{self.y}]\n [{self.n * self.u}]]"

    def __eq__(self, other):
        return (
            np.all(self.y == other.y)
            and np.all(self.u == other.u)
            and np.all(self.n == other.n)
            and np.all(self.λ0 == other.λ0)
        )

    # to enable (self @= other) syntax
    def __imatmul__(self, other):
        # other assumed to be ABCD
        return other @ self

    @property
    def wvl(self):
        # In case you don't have Greek installed:
        return self.λ0

class GaussianBeam(Ray):
    def __init__(
        self,
        *,
        λ0=532e-9,
        n=1,
        sign=1,
        q=None,
        R=None,
        w=None,
        z=None,
        zR=None,
    ):
        """Creates a GaussianBeam class. Pick any two beam 
        parameters and provide them as keyword arguments. 
        Arguments are always physical values (e.g. z is 
        distance as measured, not air-equivalent thickness 
        and q is z + 1j * zR, not reduced q).

        The reduced q value is what can be used for ABCD 
        calculations. While this is used internally, the 
        properties of GaussianBeam always return physical 
        values (including the non-reduced q)."""

        # As a convenience:
        if R == 0:
            R = math.inf

        if w == 0 or zR == 0:
            raise ValueError("Invalid argument w or zR")

        # Conditionals structured to allow values to be zero
        if q is not None:
            qr = q / n
        elif R is not None and w is not None:
            qr = equations.q_from_R_w(R, w, λ0, n)
        elif z is not None and zR is not None:
            qr = equations.q_from_z_zR(z, zR, n)
        elif R is not None and z is not None:
            qr = equations.q_from_R_z(R, z, n)
        elif R is not None and zR is not None:
            qr = equations.q_from_R_zR(R, zR, n, sign)
        elif w is not None and z is not None:
            qr = equations.q_from_w_z(w, z, λ0, n, sign)
        elif w is not None and zR is not None:
            qr = equations.q_from_w_zR(w, zR, λ0, n, sign)
        else:
            raise ValueError("No valid configuration for beam found")
        super().__init__(1, 1 / (n * qr), n, λ0)

    def __repr__(self):
        return (
            f"R & w: {self.R * 1e3:g} mm & {self.w * 1e3:g} mm\n"
            f"z & zR: {self.z * 1e3:g} mm & {self.zR * 1e3:g} mm\n"
            f"w0 & θ: {self.w0 * 1e3:g} mm & {self.θ * 1e3:g} mrad\n"
            f"q: {self.q * 1e3:g} mm"
        )

    @property
    def q(self):
        return self.y / self.u

    @property
    def R(self):
        try:
            real = (1 / self.q).real
            return 1 / real
        except ZeroDivisionError:
            return math.inf

    @property
    def w(self):
        imag = (self.n / self.q).imag
        return (-math.pi / self.λ0 * imag) ** (-0.5)

    @property
    def z(self):
        return self.q.real

    @property
    def zR(self):
        return self.q.imag

    @property
    def w0(self):
        return (self.λ0 * self.zR / (self.n * math.pi)) ** 0.5

    @property
    def θ(self):
        return self.w0 / self.zR

    @property
    def div(self):
        # In case you don't have Greek installed:
        return self.θ




import math

"""Docstring for the equations.py module.

This module contains a small collection of optical equations
to demonstrate some possible Python implementations.
"""


def rayleigh_resolution(λ, fno):
    """Calculates the image-space resolution of a lens
    using the Rayleigh criterion.

    Parameters
    ----------
    λ : float
        Vacuum wavelength in µm.
    fno : float
        f/# of the lens.

    Returns
    -------
    Z : float
        Minimum resolved distance at image in µm

    References
    ----------
    W. Smith, Modern Optical Engineering, 4th ed, p. 193.
    """
    Z = 1.22 * λ * fno
    return Z


def diffraction_angle(λ, period, angle_in=0, order=-1, sign=1):
    """Calculates the diffraction angle from a grating.

    Parameters
    ----------
    λ : float
        Vacuum wavelength in µm.
    period : float
        Grating period in μm.
    angle_in : float
        Angle of incident wave in degrees.
    order : float
        Diffraction order of interest.
    sign : float
        +1 selects the convention sin(θ1) - sin(θ2),
        whereas -1 is for sin(θ1) + sin(θ2)

    Returns
    -------
    angle_out : float
        Angle of diffracted wave in degrees.

    References
    ----------
    B. E. A. Saleh and M. C. Teich, Fundamentals of Photonics, p. 56.
    """
    sign = 1 if sign >= 0 else -1
    S1 = math.sin(math.radians(angle_in))
    return math.degrees(math.asin(order * λ / period + sign * S1))


def fiber_coupling_efficiency(
    λ,
    incoming_waist,
    fiber_waist,
    transverse=0,
    longitudinal=0,
    angular=0,
    n=1,
):
    """Calculates fiber coupling efficiency assuming
    incoming mode and fiber mode are Gaussians.

    Parameters
    ----------
    incoming_waist : float
        1/e^2 irradiance radius of incoming mode at
        focus in µm.
    fiber_waist : float
        1/e^2 irradiance radius of best-fit Gaussian
        to fiber mode in µm.
    transverse : float
        Transverse distance between incoming and fiber
        focii in µm.
    longitudinal : float
        Longitudinal distance between incoming and fiber
        focii in µm.
    angular : float
        Angle between propagation directions of incoming
        and fiber focii in degrees.
    λ : float
        Vacuum wavelength in µm.
    n : float
        Index of material surrounding fiber, defaults to air.

    Returns
    -------
    η : float
        Coupling efficiency ranging from 0 to 1.

    References
    ----------
    F. C. Allard, Fiber Optics Handbook for Engineers
    and Scientists, pp. 3.12
    """
    k = math.tau * n / λ
    A = (k * incoming_waist) ** 2 / 2
    D = (fiber_waist / incoming_waist) ** 2
    F = 2 * transverse / (k * incoming_waist**2)
    G = 2 * longitudinal / (k * incoming_waist**2)
    B = G**2 + (D + 1) ** 2
    C = (
        (D + 1) * F**2
        + 2 * D * F * G * math.sin(math.radians(angular))
        + D * (G**2 + D + 1) * math.sin(math.radians(angular)) ** 2
    )
    η = 4 * D / B * math.exp(-A * C / B)
    return η


def critical_angle(index):
    """Calculates the angle for total internal reflection
    inside a glass suspended in air.

    Parameters
    ----------
    index : float
        Index of refraction of glass relative to air.

    Returns
    -------
    angle : float
        Critical angle in degrees.

    References
    ----------
    F. C. Allard, Fiber Optics Handbook for Engineers
    and Scientists, pp. 1.3
    """
    angle = math.degrees(math.asin(1 / index))
    return angle


def NA_from_indices(core_index, cladding_index):
    """Calculates numerical aperture of a step-index fiber
    using core and cladding indices.

    Parameters
    ----------
    core_index : float
        Index of refraction of fiber core.
    cladding_index : float
        Index of refraction of fiber cladding.

    Returns
    -------
    NA : float
        Numerical aperture of fiber based on critical
        angle between fiber indices.

    References
    ----------
    F. C. Allard, Fiber Optics Handbook for Engineers
    and Scientists, pp. 1.3
    """
    NA = (core_index**2 - cladding_index**2) ** 0.5
    return NA


def best_fit_waist(core_index, cladding_index, core_radius, λ):
    """Calculates 1/e^2 irradiance radius of a Gaussian
    beam best matched to the real fiber mode given by
    `core_index,` `cladding_index,` and `wavelength.`

    Parameters
    ----------
    core_index : float
        Index of refraction of fiber core.
    cladding_index : float
        Index of refraction of fiber cladding.
    core_radius : float
        Radius of the fiber core in µm.
    λ : float
        Vacuum wavelength in µm.

    Returns
    -------
    w : float
        1/e^2 irradiance radius of best-fit Gaussian in µm.

    References
    ----------
    F. C. Allard, Fiber Optics Handbook for Engineers
    and Scientists, pp. 3.12
    """
    V = math.tau * core_radius / λ * (core_index**2 - cladding_index**2) ** 0.5
    w = core_radius * (0.65 + 1.619 / V**1.5 + 2.879 / V**6)
    return w


def fresnel_reflection(n1=1, n2=1.5, AOI=0, S_or_P="P"):
    """Calculates the field reflection coefficient ρ
    for an interface between materials with indices n1 and n2.
    For the reflected power, calcluate ρ**2. AOI is in degrees.
    S refers to polarization perpendicular to the plane of incidence.
    P refers to polarization lying inside the plane of incidence.
    Using the form from Electromagnetic Waves and Antennas by S. Orfanidis where
    the equation is recast into only n1, n2, and the angle of incidence"""
    AOI = math.radians(AOI)
    if S_or_P.lower() == "p":
        return (math.sqrt((n2 / n1) ** 2 - math.sin(AOI) ** 2) - (n2 / n1) ** 2 * math.cos(AOI)) / (
            math.sqrt((n2 / n1) ** 2 - math.sin(AOI) ** 2) + (n2 / n1) ** 2 * math.cos(AOI)
        )
    elif S_or_P.lower() == "s":
        return (math.cos(AOI) - math.sqrt((n2 / n1) ** 2 - math.sin(AOI) ** 2)) / (
            math.cos(AOI) + math.sqrt((n2 / n1) ** 2 - math.sin(AOI) ** 2)
        )


# TODO: consider error on w==0
def q_from_R_w(R, w, λ0=532e-9, n=1):
    """Calculates the reduced q value of a Gaussian beam
    when given radius R and size w"""
    return 1 / (n / R - 1j * λ0 / (math.pi * w**2))


# TODO: consider error on zR==0
def q_from_z_zR(z, zR, n=1):
    """Calculates the reduced q value of a Gaussian beam
    when given position z and Rayleigh range zR"""
    return z / n + 1j * zR / n


def q_from_R_z(R, z, n=1):
    """Calculates the reduced q value of a Gaussian beam when given
    radius R and position z based on the equation for R(z)"""
    if z == 0 or R == math.inf:
        raise ValueError("Cannot have z == 0 or R == infinity " "when specifying R and z")
    zR = ((R - z) * z) ** 0.5
    if isinstance(zR, complex) or zR == 0:
        raise ValueError("Must have |z| < |R|")
    return (z + 1j * zR) / n


# TODO: consider error on zR==0
def q_from_R_zR(R, zR, n=1, sign=1):
    """Calculates the reduced q value of a Gaussian beam when given
    radius R and Rayleigh range zR based on the equation for R(z) by
    solving 0 == z**2 - R*z + zR**2.
    sign == 1 corresponds to z such that R is in the near-field.
    sign == -1 corresponds to z such that R is in the far-field"""
    sign = 1 if sign >= 0 else -1
    # Flipping sign so this and the w & z equation match:
    z = (R - sign * (R**2 - 4 * zR**2) ** 0.5) / 2
    if isinstance(z, complex):
        raise ValueError("Must have 2 * zR <= R")
    return (z + 1j * zR) / n


# TODO: consider error on w == 0
def q_from_w_z(w, z, λ0=532e-9, n=1, sign=1):
    """Calculates the reduced q value of a Gaussian beam when given
    size w and position z based on the equation for w(z) by
    solving 0 == zR**2 - n*pi*w**2/λ0 * zR + z**2
    sign == 1 corresponds to zR such that w is in the near-field.
    sign == -1 corresponds to zR such that w is in the far-field"""
    b = -n * math.pi * w**2 / λ0
    discrim = b**2 - 4 * z**2
    sign = 1 if sign >= 0 else -1
    zR = (-b + sign * discrim**0.5) / 2
    if isinstance(zR, complex):
        raise ValueError("Must have |z| <= |n * π * w**2 / λ0|")
    return (z + 1j * zR) / n


# TODO: consider error on w or zR == 0
def q_from_w_zR(w, zR, λ0=532e-9, n=1, sign=1):
    """Calculates the reduced q value of a Gaussian beam when given
    size w and Rayleigh range zR based on equation for w(z)
    sign == 1 corresponds to z such that the beam is
        leaving the waist.
    sign == -1 corresponds to z such that the beam is
        approaching the waist"""
    sign = 1 if sign >= 0 else -1
    z = sign * ((n * math.pi * w**2 / λ0 - zR) * zR) ** 0.5
    if isinstance(z, complex):
        raise ValueError("Must have zR <= n * π * w**2 / λ0")
    return (z + 1j * zR) / n


def achromatic_doublet(EFL, na, va, nb, vb):
    """Given effective focal length and refractive index and Abbe V# of materials A and B,
    calculate the four necessary curvatures for a doublet with no spherical aberration,
    no coma, and no axial color. Uses thin-lens G-sums. Based on the procedure from
    Lens Design by Milton Laikin."""
    Fa = (va - vb) * EFL / va
    Fb = (vb - va) * EFL / vb
    Ca = 1 / (Fa * (na - 1))
    Cb = 1 / (Fb * (nb - 1))

    H = G8(nb) * Cb**2 - G8(na) * Ca**2 - G7(nb) * Cb / EFL
    I = G5(na) * Ca / 4
    K = G5(nb) * Cb / 4
    A = G1(na) * Ca**3 + G1(nb) * Cb**3 - G3(nb) * Cb**2 / EFL + G6(nb) * Cb / EFL**2
    B = -G2(na) * Ca**2
    E = G4(na) * Ca
    J = G4(nb) * Cb
    D = G2(nb) * Cb**2 - G5(nb) * Cb / EFL
    P = A + H * (J * H / K - D) / K
    Q = B + I * (2 * J * H / K - D) / K
    R = E + J * (I / K) ** 2
    root = Q**2 - 4 * P * R
    if root < 0:
        return

    s1 = {"C1": (-Q + math.sqrt(root)) / (2 * R)}
    s1["C4"] = -(H + I * s1["C1"]) / K
    s1["C2"] = s1["C1"] - Ca
    s1["C3"] = Cb + s1["C4"]
    sols = [s1]
    if root == 0:
        return sols

    s2 = {"C1": (-Q - math.sqrt(root)) / (2 * R)}
    s2["C4"] = -(H + I * s2["C1"]) / K
    s2["C2"] = s2["C1"] - Ca
    s2["C3"] = Cb + s2["C4"]
    sols.append(s2)
    return sols


def G1(n):
    return n**2 * (n - 1) / 2


def G2(n):
    return (2 * n + 1) * (n - 1) / 2


def G3(n):
    return (3 * n + 1) * (n - 1) / 2


def G4(n):
    return (n + 2) * (n - 1) / (2 * n)


def G5(n):
    return 2 * (n**2 - 1) / n


def G6(n):
    return (3 * n + 2) * (n - 1) / (2 * n)


def G7(n):
    return (2 * n + 1) * (n - 1) / (2 * n)


def G8(n):
    return n * (n - 1) / 2



"""These tests focus on testing abcd.ABCD"""
import math

import pytest

from paraxial import sources, surfaces


def test_zero_handling():
    identity = surfaces.ABCD()
    assert identity.C == 0
    assert identity.f1 == math.inf


def test_abcd_equality():
    identity_1 = surfaces.ABCD()
    identity_2 = surfaces.ABCD()
    other = surfaces.ABCD(B=1)
    assert identity_2 == identity_2
    assert identity_1 != other


def test_matrix_multiplication():
    identity = surfaces.ABCD()
    mtx1 = surfaces.ABCD(1, 0, -1 / 10e-3, 1)
    mtx2 = surfaces.ABCD(1, 5e-3, 0, 1)
    assert identity == (identity @ identity)
    assert mtx1 @ mtx2 != mtx2 @ mtx1


def test_matrix_multiplication_inplace():
    identity = surfaces.ABCD()
    mtx1 = surfaces.ABCD(1, 0, -1 / 10e-3, 1)
    mtx2 = surfaces.ABCD(1, 5e-3, 0, 1)
    # Logic is: add an mtx1, then an mtx2, etc...
    identity @= mtx1
    identity @= mtx2
    assert identity == (mtx2 @ mtx1)


def test_types():
    mat = surfaces.ABCD()
    assert isinstance(mat, surfaces.ABCD)


def test_ABCD_case_A0():
    # A == 0 --> inf-to-focus
    ray_set = [sources.Ray(n / 10) for n in range(11)]
    f = 10e-3
    mtx = surfaces.ABCD(B=f) @ surfaces.ABCD(C=-1 / f)
    assert mtx.A == pytest.approx(0)
    for ray in ray_set:
        output = mtx @ ray
        assert output.y == pytest.approx(0)

    ray_set = [sources.Ray(n / 10, 0.1) for n in range(11)]
    for ray in ray_set:
        output = mtx @ ray
        assert output.y == pytest.approx(0.1 * f)


def test_ABCD_case_B0():
    # B == 0 --> finite conjugate imaging
    ray_set = [sources.Ray(0, n / 100) for n in range(11)]
    f = 10e-3
    mtx = (
        surfaces.ABCD(B=2 * f) @ surfaces.ABCD(C=-1 / f) @ surfaces.ABCD(B=2 * f)
    )
    assert mtx.B == pytest.approx(0)
    for ray in ray_set:
        output = mtx @ ray
        assert output.y == pytest.approx(ray.y)
        assert output.u == pytest.approx(-ray.u)

    ray_set = [sources.Ray(1, n / 100) for n in range(11)]
    for ray in ray_set:
        output = mtx @ ray
        assert output.y == pytest.approx(-ray.y)


def test_ABCD_case_C0():
    # C == 0 --> afocal
    ray_set = [sources.Ray(n / 10) for n in range(11)]
    f = 10e-3
    mtx = (
        surfaces.ABCD(C=-1 / (2 * f))
        @ surfaces.ABCD(B=3 * f)
        @ surfaces.ABCD(C=-1 / f)
    )
    assert mtx.C == pytest.approx(0)
    for ray in ray_set:
        output = mtx @ ray
        assert output.y == pytest.approx(-2 * ray.y)

    ray_set = [sources.Ray(n / 10, 0.1) for n in range(11)]
    for ray in ray_set:
        output = mtx @ ray
        assert output.u == pytest.approx(-ray.u / 2)


def test_ABCD_case_D0():
    # D == 0 --> collimator
    ray_set = [sources.Ray(0, n / 10) for n in range(11)]
    f = 10e-3
    mtx = surfaces.ABCD(C=-1 / f) @ surfaces.ABCD(B=f)
    assert mtx.D == pytest.approx(0)
    for ray in ray_set:
        output = mtx @ ray
        assert output.u == pytest.approx(0)
    ray_set = [sources.Ray(1, n / 10) for n in range(11)]
    for ray in ray_set:
        output = mtx @ ray
        assert output.u == pytest.approx(-1 / f)

# TODO: add a few more pedantic and hard-coded edge cases
# e.g. displacement and matrix multiplication
def test_augmented_matrix():
    perturb = -0.1
    n = 1.5
    f = 100e-3
    axis = sources.Ray()
    marginal = sources.Ray(0, 0.1)
    chief = sources.Ray(1, 0)

    offset = surfaces.ThinLens(f, disp=perturb)
    tilt = surfaces.Refraction(math.inf, 1, n, tilt=perturb)

    for ray in (axis, marginal, chief):
        before = ray

        after = offset @ ray
        assert after.y == before.y
        assert after.u == before.u + perturb / f - before.y / f

        after = tilt @ ray
        assert after.y == before.y
        assert after.u == (before.u + perturb * (n - 1)) / n


def test_explicit_thin_lens():
    focal_length = 25
    lens = surfaces.ABCD(1, 0, -1 / focal_length, 1)

    assert lens.f1 == focal_length  # focal length calculation
    assert lens.f1 == lens.f2  # lens is in air
    assert lens.P1 == lens.P2  # principal planes superimposed
    assert lens.F1 == -lens.F2  # required by planes and fs
    assert lens.N1 == -lens.N2


def test_curved_face():
    radius = 50
    index = 1.5
    face = surfaces.ABCD(1, 0, -(index - 1) / radius, 1, n2=index)

    assert face.P1 == face.P2  # principal planes superimposed
    assert face.n1 == 1  # in air
    assert face.n2 == index  # glass-side
    assert face.f1 == radius / (index - 1)  # air-side
    assert face.f2 == radius / (index - 1) * index  # glass-side



# TODO: check temperature validity ranges


def air(λ, T=20, P=101325):
    """Calculates refractive index of air.

    Parameters
    ----------
    λ : float
        Vacuum wavelength in µm.
    T : float
        Temperature in °C.
    P : float
        Presure in Pa.

    Returns
    -------
    n : float
        Refractive index of material at λ, T, and P.

    References
    ----------
    K. Schwertz and J. H. Burge, Field Guide to Optomechanical Design and Analysis, p. 102.
    W. J. Smith, Modern Optical Engineering, 4th Ed, p. 4.
    """
    ns = (
        8342.54
        + 2406147 / (130 - 1 / λ ** 2)
        + 15998 / (38.9 - 1 / λ ** 2)
    ) * 1e-8

    n = 1 + P * ns / 96095.43 * (
        1 + 1e-8 * (0.601 - 9.72e-3 * T) * P
    ) / (1 + 3.661e-3 * T)
    return n


def nbk7(λ, T=20, P=101325):
    """Calculates refractive index of Schott N-BK7 in air.

    Parameters
    ----------
    λ : float
        Wavelength in µm.
        Valid for 365 nm to 1.06 μm
    T : float
        Temperature of glass in °C.
    P : float
        Pressure of surrounding air in Pa.

    Returns
    -------
    n : float
        Refractive index of material at λ.

    References
    ----------
    https://shop.schott.com/advanced_optics/en/SCHOTT-N-BK7/c/glass-SCHOTT%20N-BK7%C2%AE
    """

    # Absolute index:
    B1 = 1.03961212
    B2 = 0.231792344
    B3 = 1.01046945
    C1 = 0.00600069867
    C2 = 0.0200179144
    C3 = 103.560653
    n = (
        1
        + B1 * λ ** 2 / (λ ** 2 - C1)
        + B2 * λ ** 2 / (λ ** 2 - C2)
        + B3 * λ ** 2 / (λ ** 2 - C3)
    ) ** 0.5

    # Temperature dependence:
    D0 = 1.86e-6
    D1 = 1.31e-8
    D2 = -1.37e-11
    E0 = 4.34e-7
    E1 = 6.27e-10
    λtk = 0.17  # in μm
    ΔT = T - 20
    Δn = (
        (n ** 2 - 1)
        / (2 * n)
        * (
            D0 * ΔT
            + D1 * ΔT ** 2
            + D2 * ΔT ** 3
            + (E0 * ΔT + E1 * ΔT ** 2) / (λ ** 2 - λtk ** 2)
        )
    )
    n = n + Δn

    # In air:
    n = n / air(λ, T, P)
    return n


def nsf5(λ, T=20, P=101325):
    """Calculates refractive index of Schott N-SF5 in air.

    Parameters
    ----------
    λ : float
        Wavelength in µm.
        Valid for 405 nm to 2.326 μm
    T : float
        Temperature of glass in °C.
    P : float
        Pressure of surrounding air in Pa.

    Returns
    -------
    n : float
        Refractive index of material at λ.

    References
    ----------
    https://www.schott.com/shop/advanced-optics/en/Optical-Glass/N-SF5/c/glass-N-SF5
    """

    # Absolute index:
    B1 = 1.52481889
    B2 = 0.187085527
    B3 = 1.427290150
    C1 = 0.01125475600
    C2 = 0.0588995392
    C3 = 129.1416750
    n = (
        1
        + B1 * λ ** 2 / (λ ** 2 - C1)
        + B2 * λ ** 2 / (λ ** 2 - C2)
        + B3 * λ ** 2 / (λ ** 2 - C3)
    ) ** 0.5

    # Temperature dependence:
    D0 = -2.51e-07
    D1 = 1.07e-08
    D2 = -2.40e-11
    E0 = 7.85e-07
    E1 = 1.15e-09
    λtk = 0.278  # in μm
    ΔT = T - 20
    Δn = (
        (n ** 2 - 1)
        / (2 * n)
        * (
            D0 * ΔT
            + D1 * ΔT ** 2
            + D2 * ΔT ** 3
            + (E0 * ΔT + E1 * ΔT ** 2) / (λ ** 2 - λtk ** 2)
        )
    )
    n = n + Δn

    # In air:
    n = n / air(λ, T, P)
    return n


def fs7980(λ, T=22, P=101325):
    """Calculates refractive index of Corning High Purity Fused Silica 7980 in air.

    Parameters
    ----------
    λ : float
        Wavelength in µm.
        Valid for 185 nm to 1.129 μm
    T : float
        Temperature of glass in °C.
        Valid for 22 °C to 25 °C
    P : float
        Pressure of surrounding air in Pa.

    Returns
    -------
    n : float
        Refractive index of material at λ.

    References
    ----------
    https://www.corning.com/media/worldwide/csm/documents/5bf092438c5546dfa9b08e423348317b.pdf
    """

    # Absolute index:
    A0 = 2.104025406e00
    A1 = -1.456000330e-04
    A2 = -9.049135390e-03
    A3 = 8.801830992e-03
    A4 = 8.435237228e-05
    A5 = 1.681656789e-06
    A6 = -1.675425449e-08
    A7 = 8.326602461e-10

    n = (
        A0
        + A1 * λ ** 4
        + A2 * λ ** 2
        + A3 * λ ** -2
        + A4 * λ ** -4
        + A5 * λ ** -6
        + A6 * λ ** -8
        + A7 * λ ** -10
    ) ** 0.5

    # Temperature dependence:
    C0 = 9.390590
    C1 = 0.235290
    C2 = -1.318560e-03
    C3 = 3.028870e-04
    ΔT = T - 22
    Δn = (C0 + C1 / λ ** 2 + C2 / λ ** 4 + C3 / λ ** 6) * ΔT * 1e-6
    n = n + Δn

    # In air:
    n = n / air(λ, T, P)
    return n


def noa61(λ):
    """Calculates absolute refractive index of Norland Optical Adhesive 61.
    Gives absolute index, rather than relative, as it's assumed that the
    adhesive is used between glass surfaces. Calculated for T = 25 °C only.

    Parameters
    ----------
    λ : float
        Wavelength in μm.

    Returns
    -------
    n : float
        Refractive index of material at λ.

    References
    ----------
    https://www.norlandprod.com/adhesives/noa61pg2.html
    """

    # Absolute index:
    n = 1.5375 + 8290.45 / (λ * 1e3) ** 2 - 2.11046e8 / (λ * 1e3) ** 4
    return n


def znse(λ, T=20, P=101325):
    """Calculates refractive index of II-VI Zinc Selenide.

    Parameters
    ----------
    λ : float
        Wavelength in µm.
    T : float
        Temperature of glass in °C.
    P : float
        Pressure of surrounding air in Pa.

    Returns
    -------
    n : float
        Refractive index of material at λ.

    References
    ----------
    Index vs wavelength data from https://ii-vi.com/product/zinc-selenide-znse/
    Coefficients calculated using findcurves.com for SELLMOD6 2D equation
    Data was n, fit was to n**2, so removed ()**0.5 around n below.
    """

    # Absolute index:
    A = 2.4111569588609116
    B = 0.59479976285565850
    C = -0.28953183849065445
    D = 1204.4848710547462
    E = 45.910794001839250
    n = A + B * λ ** 2 / (λ ** 2 - C ** 2) + D / (λ ** 2 - E ** 2)

    # Temperature dependence:
    dndT = 61e-6  # at 10.6 μm, higher at lower λ
    ΔT = T - 20
    Δn = dndT * ΔT
    n = n + Δn

    # In air:
    n = n / air(λ, T, P)
    return n


def mgf2(λ, T=25, P=101325):
    """Calculates refractive index of Magnesium Fluoride in air.

    Parameters
    ----------
    λ : float
        Wavelength in µm.
    T : float
        Temperature of glass in °C.
    P : float
        Pressure of surrounding air in Pa.

    Returns
    -------
    n : float
        Refractive index of material at λ.

    References
    ----------
    https://refractiveindex.info/?shelf=main&book=MgF2&page=Rodriguez-de_Marcos
    Fit using findcurves.com
    """

    # Absolute index:
    A = 1.4177428299172710e00
    B = 1.1505948761303543e-02
    C = -3.4962526545629879e-01
    D = -8.0656421284475994e-03
    E = 9.5656756857168387e-02
    n = A + B / (C - λ) + D / (E - λ)

    # In air:
    n = n / air(λ, T, P)
    return n


def tio2(λ, T=25, P=101325):
    """Calculates refractive index of Titanium Dioxide in air.

    Parameters
    ----------
    λ : float
        Wavelength in µm.
    T : float
        Temperature of glass in °C.
    P : float
        Pressure of surrounding air in Pa.

    Returns
    -------
    n : float
        Refractive index of material at λ.

    References
    ----------
    https://refractiveindex.info/?shelf=main&book=TiO2&page=Sarkar
    Fit using findcurves.com
    """

    # Absolute index:
    A = 1.9226445269428725e00
    B = 2.0802606609567842e-02
    C = 1.2005327946672779e-01
    D = -3.0426606351792251e-01
    n = A + B / λ ** 2 + C * λ ** 2 / (λ ** 2 - D ** 2)

    # In air:
    n = n / air(λ, T, P)
    return n



import math

import pytest
import numpy as np
import sympy as sm

from paraxial import sources, surfaces


def test_ray_numpy_use():
    heights = np.linspace(0, 1, 100)
    angles = np.linspace(1, 0, 100)
    rays = sources.Ray(heights, angles)
    identical_rays = sources.Ray(heights, angles)
    assert np.all(rays == identical_rays)
    
def test_ray_inplace_numpy():
    n1 = 1.5
    n2 = 2
    heights = np.linspace(0, 1, 10)
    angles = np.linspace(1, 0, 10)
    
    rays = sources.Ray(heights, angles)
    assert np.allclose(rays.u, angles)
    
    glass = surfaces.ABCD(n1=1, n2=n1)
    rays @= glass
    assert np.allclose(rays.u, angles/n1)
    
    air = surfaces.ABCD(n1=n1, n2=1)
    rays @= air
    assert np.allclose(rays.u, angles)

    glass = surfaces.ABCD(n1=1, n2=n2)
    rays @= glass
    assert np.allclose(rays.u, angles/n2)




"""These tests focus on testing children of abcd.ABCD involving
multiple ABCDs in their creation"""

import pytest

from paraxial import surfaces


def test_thick_lens():
    thick = surfaces.ThickLens(25e-3, -30e-3, 2e-3, 1.5)
    assert thick.f1 == thick.f2  # lens is in air
    assert thick.P1 != thick.P2  # principal planes NOT superimposed
    assert thick.F1 != -thick.F2  # required by planes and fs
    assert thick.N1 != -thick.N2


def test_lensmaker_equation():
    # equivalent by lensmaker's equation
    thin = surfaces.ThinLens(10e-3)
    thick = surfaces.ThickLens(10e-3, -10e-3, 0, 1.5)
    assert thin == thick




import numbers

import pytest

from paraxial import sources, surfaces


def test_ray_construction():
    r = sources.Ray()
    assert isinstance(r, sources.Ray)
    assert isinstance(r.y, numbers.Number)
    assert isinstance(r.u, numbers.Number)


def test_types():
    mat = surfaces.ABCD()
    ray = sources.Ray()
    assert isinstance(ray, sources.Ray)
    assert isinstance(mat @ ray, sources.Ray)


def test_explicit_update():
    y = 1
    u = 0.1
    f = 10e-3
    ray = sources.Ray(y, u)
    mtx = surfaces.ABCD(C=-1 / f) @ surfaces.ABCD(B=f)
    ray = mtx @ ray
    assert ray.y == pytest.approx(y + u * f)
    assert ray.u == pytest.approx(-1 / f)


def test_inplace_update():
    y = 1
    u = 0.1
    f = 10e-3
    ray = sources.Ray(y, u)
    mtx = surfaces.ABCD(C=-1 / f) @ surfaces.ABCD(B=f)
    ray @= mtx
    assert ray.y == pytest.approx(y + u * f)
    assert ray.u == pytest.approx(-1 / f)


def test_ray_angles_across_interface():
    m = 0.1
    n = 1.5

    marginal = sources.Ray(0, m)
    chief = sources.Ray(1, 0)

    assert marginal.y == 0
    assert marginal.u == m
    assert chief.y == 1
    assert chief.u == 0

    interface = surfaces.ABCD(n2=n)
    marginal @= interface
    chief @= interface
    assert marginal.y == 0
    assert marginal.u == pytest.approx(m / n)
    assert chief.y == pytest.approx(1)
    assert chief.u == pytest.approx(0)
    assert marginal.n == chief.n == n


def test_multiple_ray_updates():
    m = 0.1
    L = 50e-3
    n = 1.5
    R = L / 2

    marginal = sources.Ray(0, m)
    chief = sources.Ray(1, 0)

    fs = surfaces.ABCD(B=L)
    marginal @= fs
    chief @= fs
    assert marginal.y == pytest.approx(L * m)
    assert marginal.u == m
    assert chief.y == 1
    assert chief.u == 0

    refr = surfaces.ABCD(C=-(n - 1) / R, n2=n)
    marginal @= refr
    chief @= refr
    assert marginal.y == pytest.approx(L * m)
    assert marginal.u == pytest.approx(0)
    assert chief.y == pytest.approx(1)
    assert chief.u == pytest.approx(-(n - 1) / (n * R))
    assert marginal.n == chief.n == n

    L_focus = n * R / (n - 1)  # solved (L/n mtx)(y nu) for y == 0
    fs = surfaces.ABCD(B=L_focus / n, n1=n, n2=n)
    marginal @= fs
    chief @= fs
    assert marginal.y == pytest.approx(L * m)
    assert marginal.u == pytest.approx(0)
    assert chief.y == pytest.approx(0)
    assert chief.u == pytest.approx(-(n - 1) / (n * R))
    assert marginal.n == chief.n == n

    to_air = surfaces.ABCD(n1=n, n2=1)
    marginal @= to_air
    chief @= to_air
    assert marginal.y == pytest.approx(L * m)
    assert marginal.u == pytest.approx(0)
    assert chief.y == pytest.approx(0)
    assert chief.u == pytest.approx(-(n - 1) / R)
    assert marginal.n == chief.n == 1




import math
import random

import pytest

from paraxial import sources, surfaces


def test_types():
    mat = surfaces.ABCD()
    gb = sources.GaussianBeam(z=10e-3, zR=1e-3)
    assert isinstance(mat, surfaces.ABCD)
    assert isinstance(gb, sources.Ray)
    assert isinstance(gb, sources.GaussianBeam)
    assert isinstance(mat @ gb, sources.GaussianBeam)


def test_gb_attributes():
    λ0 = 500e-9
    z = 10e-3
    zR = 1e-3
    gb = sources.GaussianBeam(λ0=λ0, z=z, zR=zR)
    assert gb.n == 1
    assert gb.λ0 == λ0
    assert gb.y == 1
    assert gb.u == pytest.approx(1 / (z + 1j * zR))


def test_gb_equality():
    g1 = sources.GaussianBeam(R=100e-3, w=1e-3)
    g2 = sources.GaussianBeam(R=100e-3, w=1e-3)
    assert g1 == g2


def test_arg_error_catching():
    with pytest.raises(ValueError):
        sources.GaussianBeam(z=1, zR=0)
    with pytest.raises(ValueError):
        sources.GaussianBeam(R=1, w=0)


def test_gb_from_q():
    gb = sources.GaussianBeam(q=10e-3 + 1e-3j, n=1.5)
    assert gb.q == pytest.approx(10e-3 + 1e-3j)


def test_gb_from_R_w():
    gb = sources.GaussianBeam(R=100e-3, w=1e-3, n=1.5)
    assert gb.R == 100e-3
    assert gb.w == pytest.approx(1e-3)


def test_gb_from_z_zR():
    gb = sources.GaussianBeam(z=10e-3, zR=1e-3, n=1.5)
    assert gb.z == 10e-3
    assert gb.zR == pytest.approx(1e-3)


def test_gb_from_R_z():
    gb = sources.GaussianBeam(R=100e-3, z=10e-3, n=1.5)
    assert gb.R == pytest.approx(100e-3)
    assert gb.z == pytest.approx(10e-3)


def test_gb_from_R_zR():
    gb = sources.GaussianBeam(R=100e-3, zR=1e-3, n=1.5)
    assert gb.R == pytest.approx(100e-3)
    assert gb.zR == pytest.approx(1e-3)


def test_gb_from_w_z():
    gb = sources.GaussianBeam(w=1e-3, z=10e-3, n=1.5)
    assert gb.w == pytest.approx(1e-3)
    assert gb.z == pytest.approx(10e-3)

    gb = sources.GaussianBeam(w=1e-3, z=0)
    assert gb.w0 == pytest.approx(1e-3)
    # from Edmund Optics GB calculator:
    assert gb.θ == pytest.approx(0.16934e-3, abs=1e-7)
    assert gb.R == pytest.approx(math.inf)


def test_gb_from_w_zR():
    gb = sources.GaussianBeam(w=1e-3, zR=1e-3, n=1.5)
    assert gb.w == pytest.approx(1e-3)
    assert gb.zR == pytest.approx(1e-3)


def test_convenience():
    gb = sources.GaussianBeam(R=0, w=1e-3)
    assert gb.R == math.inf


def test_flat_interface():
    n1 = 1 + random.random() * 3
    n2 = 1 + random.random() * 3
    z = 10e-3
    zR = 1e-3

    gb1 = sources.GaussianBeam(z=z, zR=zR, n=n1)
    glass = surfaces.ABCD(n1=n1, n2=n2)
    gb2 = glass @ gb1

    # attributes:
    assert gb1.n == n1
    assert gb2.n == n2
    assert gb1.λ0 == gb2.λ0

    # properties:
    assert gb2.R == pytest.approx(n2 / n1 * gb1.R)
    assert gb2.w == pytest.approx(gb1.w)

    assert gb2.z == pytest.approx(n2 / n1 * gb1.z)
    assert gb2.zR == pytest.approx(n2 / n1 * gb1.zR)

    assert gb2.w0 == pytest.approx(gb1.w0)
    assert gb2.div == pytest.approx(
        n1 / n2 * gb1.div
    )


def test_free_space():
    # before & after free space
    n = 1 + random.random() * 3
    z = 10e-3
    zR = 1e-3
    L = 100e-3

    gb1 = sources.GaussianBeam(z=z, zR=zR, n=n)
    fs = surfaces.ABCD(B=L / n, n1=n, n2=n)
    gb2 = fs @ gb1

    # attributes:
    assert gb1.n == gb2.n == n
    assert gb1.λ0 == gb2.λ0

    # properties:
    assert gb2.z == pytest.approx(
        gb1.z + L
    )  # no 1/n because same media
    assert gb2.zR == pytest.approx(gb1.zR)

    assert gb2.w0 == pytest.approx(gb1.w0)
    assert gb2.div == pytest.approx(gb1.div)


def test_prop_to_focus_air():
    # verify distance to focus is physical, not reduced
    z = -10e-3
    zR = 1e-3
    L = -z

    gb1 = sources.GaussianBeam(z=z, zR=zR)
    glass = surfaces.ABCD(B=L)
    gb2 = glass @ gb1

    # attributes:
    assert gb1.λ0 == gb2.λ0
    assert gb1.n == gb2.n

    # properties:
    assert abs(gb2.R) >= 1e9
    assert gb2.w == pytest.approx(gb1.w0)

    assert gb2.z == pytest.approx(0)
    assert gb2.zR == pytest.approx(gb1.zR)


def test_prop_to_focus_glass():
    # verify distance to focus is physical, not reduced
    n2 = 1.5
    z = -10e-3
    zR = 1e-3
    L = -z * n2  # physical prop distance is 15 mm

    gb1 = sources.GaussianBeam(z=z, zR=zR)
    glass = surfaces.ABCD(B=L / n2, n2=n2)  # need L/n
    gb2 = glass @ gb1

    # attributes:
    assert gb2.λ0 == gb1.λ0

    # properties:
    assert abs(gb2.R) >= 1e12
    assert gb2.w == pytest.approx(gb1.w0)

    assert gb2.z == pytest.approx(0)
    assert gb2.zR == pytest.approx(n2 * gb1.zR)


def test_refraction():
    # before & after refraction (e.g. updating n)
    n = 1 + random.random() * 3
    z = 10e-3
    zR = 1e-3
    R = 100e-3

    gb1 = sources.GaussianBeam(z=z, zR=zR)
    refr = surfaces.ABCD(C=-(n - 1) / R, n2=n)
    gb2 = refr @ gb1

    # attributes:
    assert gb1.λ0 == gb2.λ0

    # properties:
    assert gb2.n / gb2.R == pytest.approx(gb1.n / gb1.R - (n - 1) / R)
    assert gb2.w == pytest.approx(gb1.w)


def test_init_in_material():
    # starting in-material vs out
    n1 = 1  # 1 + random.random() * 3
    n2 = 1 + random.random() * 3
    z = 10e-3
    zR = 1e-3

    gb1 = sources.GaussianBeam(z=z, zR=zR, n=n1)
    gb2 = sources.GaussianBeam(z=z, zR=zR, n=n2)

    assert gb1.n == n1
    assert gb2.n == n2

    assert gb2.R == pytest.approx(gb1.R)
    assert gb2.z == pytest.approx(gb1.z)
    assert gb2.zR == pytest.approx(gb1.zR)

    assert gb2.w0 == pytest.approx(gb1.w0 / n2 ** 0.5)


def test_in_place_update():
    z = 10e-3
    zR = 1e-3
    gb1 = sources.GaussianBeam(z=z, zR=zR)
    glass = surfaces.ABCD(B=10e-3)
    gb2 = glass @ gb1
    gb1 @= glass
    assert gb1 == gb2


def test_self_overwrite_update():
    z = 10e-3
    zR = 1e-3
    gb1 = sources.GaussianBeam(z=z, zR=zR)
    glass = surfaces.ABCD(B=10e-3)
    gb2 = glass @ gb1
    gb1 = glass @ gb1
    assert gb1 == gb2



import math

import pytest

from paraxial import equations


def test_rayleigh_resolution():
    λ = 0.532  # μm
    fno = 10
    ref_value = 1.22 * λ * fno  # μm
    calc_value = equations.rayleigh_resolution(λ, fno)
    assert calc_value == pytest.approx(ref_value)


def test_diffraction_angle_normal():
    λ = 0.532  # μm
    Δ = 1  # μm
    ref_value = -32.140687
    calc_value = equations.diffraction_angle(λ, Δ)
    assert calc_value == pytest.approx(ref_value)


def test_diffraction_angle_angled():
    λ = 0.532  # μm
    Δ = 1  # μm
    angle = 10
    ref_value = -20.99901
    calc_value = equations.diffraction_angle(λ, Δ, angle)
    assert calc_value == pytest.approx(ref_value)


def test_diffraction_angle_order():
    λ = 0.532  # μm
    Δ = 1  # μm
    angle = 10
    order = -2
    ref_value = -62.91749
    calc_value = equations.diffraction_angle(λ, Δ, angle, order)
    assert calc_value == pytest.approx(ref_value)


# TODO: check for different index
def test_fiber_coupling():
    # in μm
    λ = 1.55
    w1 = 5
    w2 = 5
    ref_value = 1
    calc_value = equations.fiber_coupling_efficiency(λ, w1, w2)
    assert calc_value == pytest.approx(ref_value)


def test_fiber_coupling_extremes():
    # in μm
    λ = 1.55
    w1 = 5
    w2 = 5
    ref_value = 0
    calc_value = equations.fiber_coupling_efficiency(
        λ, w1, w2, transverse=1e3
    )
    assert calc_value == pytest.approx(ref_value)

    calc_value = equations.fiber_coupling_efficiency(
        λ, w1, w2, longitudinal=1e12
    )
    assert calc_value == pytest.approx(ref_value)

    calc_value = equations.fiber_coupling_efficiency(
        λ, w1, w2, angular=90
    )
    assert calc_value == pytest.approx(ref_value)


def test_critical_angle():
    ref_value = 30
    calc_value = equations.critical_angle(2)
    assert calc_value == pytest.approx(ref_value)


def test_NA_from_indices():
    from paraxial import materials
    test_NA = 0.1
    n_core = materials.fs7980(0.532)
    n_cladding = (n_core**2 - test_NA**2)**0.5
    NA = equations.NA_from_indices(n_core, n_cladding)
    assert NA == pytest.approx(test_NA)



def test_fresnel_reflection():
    n = 1.5
    rho = equations.fresnel_reflection(1, n)
    assert rho**2 == pytest.approx(0.04)

    θ1 = 10
    θ2 = math.degrees(math.asin(math.sin( math.radians(θ1) ) / n))  # Snel's law

    rho = equations.fresnel_reflection(1, 1.5, θ1, S_or_P='P')
    nt1 = 1 / math.cos(math.radians(θ1))
    nt2 = n / math.cos(math.radians(θ2))
    rho_explicit = (nt1 - nt2) / (nt1 + nt2)
    assert rho**2 == pytest.approx(rho_explicit**2)

    rho = equations.fresnel_reflection(1, 1.5, θ1, S_or_P='S')
    nt1 = math.cos(math.radians(θ1))
    nt2 = n * math.cos(math.radians(θ2))
    rho_explicit = (nt1 - nt2) / (nt1 + nt2)
    assert rho**2 == pytest.approx(rho_explicit**2)


# tests for each q_from_X_X function
def test_q_from_Rw():
    R = 10e-3
    w = 1e-3
    λ0 = 532e-9
    n = 1.5
    q = equations.q_from_R_w(R, w, λ0=λ0, n=n)
    assert 1 / ((1 / q).real / n) == pytest.approx(R)
    assert 1 / math.sqrt(
        -(1 / q).imag * math.pi / λ0
    ) == pytest.approx(w)


def test_q_from_zzR():
    z = 10e-3
    zR = 1e-3
    n = 1.5
    q = equations.q_from_z_zR(z, zR, n=n)
    assert q.real * n == pytest.approx(z)
    assert q.imag * n == pytest.approx(zR)


def test_q_from_Rz():
    R = 10e-3
    z = 1e-3
    n = 1.5
    q = equations.q_from_R_z(R, z, n=n)
    assert 1 / ((1 / q).real / n) == pytest.approx(R)
    assert q.real * n == pytest.approx(z)

    with pytest.raises(ValueError):
        q = equations.q_from_R_z(R=1, z=1, n=n)
    with pytest.raises(ValueError):
        q = equations.q_from_R_z(R=1, z=2, n=n)
    with pytest.raises(ValueError):
        q = equations.q_from_R_z(R=1, z=0, n=n)


def test_q_from_RzR():
    R = 10e-3
    zR = 1e-3
    n = 1.5
    sign = 1
    q = equations.q_from_R_zR(R, zR, n=n, sign=sign)
    assert 1 / ((1 / q).real / n) == pytest.approx(R)
    assert q.imag * n == pytest.approx(zR)

    with pytest.raises(ValueError):
        q = equations.q_from_R_zR(R=1, zR=0.500001, n=n)

    z1 = q.real * n
    sign = -1
    q = equations.q_from_R_zR(R, zR, n=n, sign=sign)
    z2 = q.real * n
    assert z1 < z2


def test_q_from_wz():
    w = 1e-3
    z = 10e-3
    λ0 = 532e-9
    n = 1.5
    sign = 1
    q = equations.q_from_w_z(w, z, λ0=λ0, n=n, sign=sign)
    assert 1 / math.sqrt(
        -(1 / q).imag * math.pi / λ0
    ) == pytest.approx(w)
    assert q.real * n == pytest.approx(z)

    with pytest.raises(ValueError):
        q = equations.q_from_w_z(
            w=1, z=n * math.pi / λ0 * 1.0001, n=n
        )
    with pytest.raises(ValueError):
        q = equations.q_from_w_z(
            w=-1, z=-n * math.pi / λ0 * 1.0001, n=n
        )

    zR1 = q.imag * n
    sign = -1
    q = equations.q_from_w_z(w, z, λ0=λ0, n=n, sign=sign)
    zR2 = q.real * n
    assert zR1 > zR2


def test_q_from_wzR():
    w = 1e-3
    zR = 1e-3
    λ0 = 532e-9
    n = 1.5
    sign = 1
    q = equations.q_from_w_zR(w, zR, λ0=λ0, n=n, sign=sign)
    assert 1 / math.sqrt(
        -(1 / q).imag * math.pi / λ0
    ) == pytest.approx(w)
    assert q.imag * n == pytest.approx(zR)

    with pytest.raises(ValueError):
        q = equations.q_from_w_zR(
            w=1, zR=n * math.pi / λ0 * 1.0001, n=n
        )

    z1 = q.real * n
    sign = -1
    q = equations.q_from_w_zR(w, zR, λ0=λ0, n=n, sign=sign)
    z2 = q.real * n
    assert z1 == -z2


def test_doublet_equation():
    from paraxial import materials
    λC = 0.6563  # µm
    λD = 0.5893  # µm
    λF = 0.4861  # µm

    na = materials.nbk7(λD)
    nb = materials.nsf5(λD)

    va = (materials.nbk7(λD) - 1) / (materials.nbk7(λF) - materials.nbk7(λC))
    vb = (materials.nsf5(λD) - 1) / (materials.nsf5(λF) - materials.nsf5(λC))

    EFL = 100e-3
    solutions = equations.achromatic_doublet(EFL, na, va, nb, vb)
    assert len(solutions[0]) == 4
    assert len(solutions[1]) == 4
    assert round(solutions[0]['C2'], 1) == round(solutions[0]['C3'], 1)




"""These tests focus on testing individual children of abcd.ABCD"""
import math

import pytest

from paraxial import sources, surfaces
from paraxial import equations


def test_transfer():
    t = 10e-3
    n = 1.5
    trans = surfaces.Transfer(t, n)
    assert trans.B == t / n


def test_refraction():
    n1 = 1
    n2 = 1.5
    R = 50
    refr = surfaces.Refraction(R, n1, n2)
    assert refr.n1 == n1
    assert refr.n2 == n2
    assert refr.f1 == R / (n2 - n1)  # air-side
    assert refr.f2 == R / (n2 - n1) * n2  # glass-side


def test_refraction_nonzero_AOI():
    R = 100e-3
    n = 1.5
    AOI = 0.1
    AOE = math.asin(math.sin(AOI) / n)
    C1 = math.cos(AOI)
    C2 = math.cos(AOE)
    M = C2 / C1

    surface = surfaces.Refraction(R, 1, n, AOI, "T")
    dne = (n * C2 - C1) / (C1 * C2)
    assert surface.A == pytest.approx(M)
    assert surface.B == 0
    assert surface.C == -dne / R
    assert surface.D == pytest.approx(1 / M)

    surface = surfaces.Refraction(R, 1, n, AOI, "S")
    dne = n * C2 - C1
    assert surface.A == 1
    assert surface.B == 0
    assert surface.C == -dne / R
    assert surface.D == 1

    surface = surfaces.Refraction(math.inf, 1, n, AOI, "T")
    assert surface.C == 0

    surface = surfaces.Refraction(math.inf, 1, n, AOI, "S")
    assert surface.C == 0


def test_mirror():
    R = 50
    refl = surfaces.Mirror(R)
    assert refl.n1 == 1
    assert refl.n2 == 1
    assert refl.A == 1
    assert refl.B == 0
    assert refl.C == 2 / R
    assert refl.D == 1


def test_mirror_nonzero_AOI():
    R = 100e-3
    AOI = 0.1

    surface = surfaces.Mirror(R, AOI=AOI, T_or_S="T")
    Re = R * math.cos(AOI)
    assert surface.A == 1
    assert surface.B == 0
    assert surface.C == 2 / Re
    assert surface.D == 1

    surface = surfaces.Mirror(R, AOI=AOI, T_or_S="S")
    Re = R / math.cos(AOI)
    assert surface.A == 1
    assert surface.B == 0
    assert surface.C == 2 / Re
    assert surface.D == 1

    surface = surfaces.Mirror(math.inf, AOI=AOI, T_or_S="T")
    assert surface.C == 0

    surface = surfaces.Mirror(math.inf, AOI=AOI, T_or_S="S")
    assert surface.C == 0


def test_duct():
    # Short duct is thin lens with B=z/n0 and C=-n2*z
    duct = surfaces.Duct(1e-9, 1, 10e9)
    approx_thin_lens = surfaces.ABCD(1, 1e-9, -1 / 100e-3, 1)
    assert duct.A == pytest.approx(approx_thin_lens.A)
    assert duct.B == pytest.approx(approx_thin_lens.B)
    assert duct.C == pytest.approx(approx_thin_lens.C)
    assert duct.D == pytest.approx(approx_thin_lens.D)


def test_grating():
    # Based on a Thorlabs grating in Littrow geometry
    λ0 = 500e-9  # blaze wavelength 500 nm
    d = 1e-3 / 1200  # 1200 lp/mm
    AOI = math.radians(17 + 27 / 60)  # blaze angle
    AOE = math.radians(equations.diffraction_angle(λ0 * 1e6, d * 1e6, math.degrees(AOI)))
    C1 = math.cos(AOI)
    C2 = math.cos(AOE)
    M = C2 / C1

    R = math.inf
    Rt = 2 * R * C1 * C2 / (C1 + C2)
    Rs = 2 * R / (C1 + C2)

    grating = surfaces.Grating(R=R, d=d, λ0=λ0, AOI=AOI, T_or_S="T")
    assert grating.A == M
    assert grating.B == 0
    assert grating.C == 0
    assert grating.D == 1 / M

    grating = surfaces.Grating(R=R, d=d, λ0=λ0, AOI=AOI, T_or_S="S")
    assert grating.A == 1
    assert grating.B == 0
    assert grating.C == 2 / Rs
    assert grating.D == 1

    R = 100e-3
    Rt = 2 * R * C1 * C2 / (C1 + C2)
    Rs = 2 * R / (C1 + C2)

    grating = surfaces.Grating(R=R, d=d, λ0=λ0, AOI=AOI, T_or_S="T")
    assert grating.A == M
    assert grating.B == 0
    assert grating.C == 2 / Rt
    assert grating.D == 1 / M

    grating = surfaces.Grating(R=R, d=d, λ0=λ0, AOI=AOI, T_or_S="S")
    assert grating.A == 1
    assert grating.B == 0
    assert grating.C == 2 / Rs
    assert grating.D == 1


def test_thin_lens():
    f = 25e-3
    thin = surfaces.ThinLens(f)
    assert thin.f1 == thin.f2 == f
    assert thin.A == thin.D == 1
    assert thin.B == 0
    assert thin.C == -1 / f



import math
import copy

import numpy as np

from . import sources

# TODO: check for non-unity determinant?
class ABCD:
    def __init__(
        self, A=1, B=0, C=0, D=1, n1=1, n2=1, disp=0, tilt=0, length=0, E=None, F=None
    ):
        self.n1 = n1
        self.n2 = n2
        self.A = A
        self.B = B
        self.C = C
        self.D = D
        self.E = (1 - A) * disp + (length - n1 * B) * tilt if E is None else E
        self.F = -C * disp + (n2 - n1 * D) * tilt if F is None else F

    def __repr__(self):
        return (
            f"[[{self.A:g} {self.B:g} {self.E:g}]\n"
            f" [{self.C:g} {self.D:g} {self.F:g}]\n"
            f" [0 0 1]]"
        )

    def __eq__(self, other):
        return (
            (self.A == other.A)
            and (self.B == other.B)
            and (self.C == other.C)
            and (self.D == other.D)
            and (self.E == other.E)
            and (self.F == other.F)
        )

    # to enable (self @ other) syntax
    def __matmul__(self, other):
        # self @ other
        if isinstance(other, ABCD):
            A = self.A * other.A + self.B * other.C
            B = self.A * other.B + self.B * other.D
            C = self.C * other.A + self.D * other.C
            D = self.C * other.B + self.D * other.D
            E = self.A * other.E + self.B * other.F + self.E
            F = self.C * other.E + self.D * other.F + self.F
            return ABCD(A=A, B=B, C=C, D=D, n1=other.n1, n2=self.n2, E=E, F=F)
        # self @ Ray
        elif isinstance(other, sources.Ray):
            y = self.A * other.y + self.B * other.n * other.u + self.E
            u = (
                self.C * other.y + self.D * other.n * other.u + self.F
            ) / self.n2

            ret = copy.deepcopy(other)  # so "other" unchanged
            ret.y = y
            ret.u = u
            ret.n = copy.deepcopy(self.n2)
            return ret

    # to enable (self @= other) syntax
    def __imatmul__(self, other):
        return other @ self

    def add_misalignments(self, disp, tilt, length):
        self.E = (1 - self.A) * disp + (length - self.n1 * self.B) * tilt 
        self.F = -self.C * disp + (self.n2 - self.n1 * self.D) * tilt

    @property
    def F1(self):
        try:
            return self.n1 * self.D / self.C
        except ZeroDivisionError:
            return -math.inf

    @property
    def P1(self):
        try:
            return self.n1 * (self.D - 1) / self.C
        except ZeroDivisionError:
            return -math.inf

    @property
    def N1(self):
        try:
            return (self.D * self.n1 - self.n2) / self.C
        except ZeroDivisionError:
            return -math.inf

    @property
    def f1(self):
        try:
            return -self.n1 / self.C
        except ZeroDivisionError:
            return math.inf

    @property
    def F2(self):
        try:
            return -self.n2 * self.A / self.C
        except ZeroDivisionError:
            return math.inf

    @property
    def P2(self):
        try:
            return self.n2 * (1 - self.A) / self.C
        except ZeroDivisionError:
            return math.inf

    @property
    def N2(self):
        try:
            return (self.n1 - self.A * self.n2) / self.C
        except ZeroDivisionError:
            return math.inf

    @property
    def f2(self):
        try:
            return -self.n2 / self.C
        except ZeroDivisionError:
            return math.inf


class Transfer(ABCD):
    def __init__(self, t, n=1, **kwargs):
        super().__init__(1, t / n, 0, 1, **kwargs)


class Refraction(ABCD):
    """ABCD matrix for refraction at dielectric interface.
    AOI is angle-of-incidence of axial ray in radians
    (e.g. coming in to lens at a non-normal angle) 
    and T_or_S specifies tangential or saggital. R is
    positive for a convex surface, typical of geometrical
    optics but counter to Siegman."""

    def __init__(self, R, n1, n2, AOI=0, T_or_S="T", **kwargs):
        AOE = math.asin(n1 * math.sin(AOI) / n2)
        C1 = math.cos(AOI)
        C2 = math.cos(AOE)

        if T_or_S.lower() == "t":
            denominator = C1 * C2
            A = C2 / C1
            D = C1 / C2
        elif T_or_S.lower() == "s":
            denominator = 1
            dne = n2 * C2 - n1 * C1
            A = D = 1
        else:
            raise ValueError(f"Unknown T_or_S: {T_or_S}")
        numerator = n2 * C2 - n1 * C1
        dne = numerator / denominator
        B = 0
        C = -dne / R
        super().__init__(A, B, C, D, n1, n2, **kwargs)


class Mirror(ABCD):
    """ABCD matrix for reflection at mirror.
    AOI is angle-of-incidence of axial ray in radians
    (e.g. coming in to mirror at a non-normal angle) 
    and T_or_S specifies tangential or saggital. R is
    positive for a convex surface, typical of geometrical
    optics but counter to Siegman."""

    def __init__(self, R, AOI=0, T_or_S="T", **kwargs):
        if T_or_S.lower() == "t":
            Re = R * math.cos(AOI)
        elif T_or_S.lower() == "s":
            Re = R / math.cos(AOI)
        else:
            raise ValueError(f"Unknown T_or_S: {T_or_S}")
        super().__init__(1, 0, 2 / Re, 1, **kwargs)


class Duct(ABCD):
    """ABCD matrix for media with radially varying index as
    n(y) = n0 - n2 * y**2 / 2. For example, a GRIN fiber or
    index variation caused by thermal lensing."""

    def __init__(self, t, n0, n2, **kwargs):
        g = (n2 / n0) ** 0.5
        A = math.cos(g * t)
        B = math.sin(g * t) / (n0 * g)
        C = -n0 * g * math.sin(g * t)
        D = math.cos(g * t)
        super().__init__(A, B, C, D, **kwargs)


class Grating(ABCD):
    """ABCD matrix for diffraction grating per Siegman. 
    Grating lines run in the x / saggital direction with 
    grating spacing d in the y / tangential direction. 
    Diffraction takes place in the YZ plane.
    Calling diffraction order m. Assumed to be reflective
    with positive R convex. sign == 1 is for the convention
    sin(θ1) - sin(θ2) whereas sign == -1 specifies
    sin(θ1) + sin(θ2)"""

    # TODO: check math for non-vacuum media
    def __init__(
        self,
        R,
        m=-1,
        d=1e-6,
        λ0=532e-9,
        AOI=0,
        T_or_S="T",
        sign=1,
        **kwargs,
    ):
        sign = 1 if sign >= 0 else -1
        AOE = math.asin(m * λ0 / d + sign * math.sin(AOI))
        C1 = math.cos(AOI)
        C2 = math.cos(AOE)

        if T_or_S.lower() == "t":
            # "Lasers" is missing the 2x here:
            Rt = 2 * R * C1 * C2 / (C1 + C2)
            M = C2 / C1
            A = M
            B = 0
            C = 2 / Rt
            D = 1 / M
        elif T_or_S.lower() == "s":
            Rs = 2 * R / (C1 + C2)
            A = D = 1
            B = 0
            C = 2 / Rs
        else:
            raise ValueError(f"Unknown T_or_S: {T_or_S}")
        super().__init__(A, B, C, D, **kwargs)


class ThinLens(ABCD):
    def __init__(self, f, **kwargs):
        super().__init__(1, 0, -1 / f, 1, **kwargs)


class ThickLens(ABCD):
    def __init__(self, R1, R2, t, n_glass, n_ambient=1, **kwargs):
        r1 = Refraction(R1, n_ambient, n_glass)
        t = Transfer(t, n_glass)
        r2 = Refraction(R2, n_glass, n_ambient)
        net = r2 @ t @ r1
        super().__init__(
            net.A, net.B, net.C, net.D, net.n1, net.n2, **kwargs
        )




import pytest

from paraxial import materials


def test_air():
    # Value from emtoolbox.nist.gov Edlen equation calculator
    assert materials.air(0.5) == pytest.approx(1.000274)


def test_nbk7():
    λ = 0.58756
    ref_index = 1.51680011
    calc_index = materials.nbk7(λ) * materials.air(λ)
    assert calc_index == pytest.approx(ref_index)

    T = 30
    heated_index = 1.51683  # unsure if in air or absolute
    assert materials.nbk7(λ, T) * materials.air(λ, T)  == pytest.approx(heated_index, abs=2e-5)


def test_nsf5():
    λ = 0.5461
    ref_index = 1.67763
    calc_index = materials.nsf5(λ) * materials.air(λ)
    assert calc_index == pytest.approx(ref_index)

    λ = 0.5461
    ref_dndT = 2e-6
    T = 30
    ΔT = T - 20
    heated_index = ref_index + ΔT * ref_dndT
    assert materials.nsf5(λ, T) * materials.air(λ, T) == pytest.approx(heated_index)


def test_fs7980():
    λ = 0.587725
    ref_index = 1.458461
    calc_index = materials.fs7980(λ) * materials.air(λ)
    assert calc_index == pytest.approx(ref_index, abs=5e-6)

    T = 32
    ΔT = T - 22
    ref_dndT = 10.1e-6
    heated_index = ref_index + ΔT * ref_dndT
    calc_index = materials.fs7980(λ, T) * materials.air(λ, T)
    assert calc_index == pytest.approx(heated_index)


# Loosened tolerance to match, perhaps fit doesn't match data superbly:
def test_noa61():
    λ = 0.5896
    ref_index = 1.5594
    assert materials.noa61(λ) == pytest.approx(ref_index, abs=2.5e-4)


# Loosened tolerance to match, perhaps fit doesn't match data superbly:
def test_znse():
    λ = 10.6
    ref_index = 2.4028
    calc_index = materials.znse(λ) * materials.air(λ)
    assert calc_index == pytest.approx(ref_index, abs=1e-4)

    T = 30
    ΔT = T - 20
    ref_dndT = 61e-6
    heated_index = ref_index + ΔT * ref_dndT
    calc_index = materials.znse(λ, T) * materials.air(λ, T)
    assert calc_index == pytest.approx(heated_index, abs=1e-4)


# Loosened tolerance to match, perhaps fit doesn't match data superbly:
def test_mgf2():
    λ = 0.5870740
    ref_index = 1.421977
    calc_index = materials.mgf2(λ) * materials.air(λ)
    assert calc_index == pytest.approx(ref_index, abs=1e-4)


# Loosened tolerance to match, perhaps fit doesn't match data superbly:
def test_tio2():
    λ = 0.587
    ref_index = 2.146858
    calc_index = materials.tio2(λ) * materials.air(λ)
    assert calc_index == pytest.approx(ref_index, abs=5e-3)
